#
# Copyright 2016-2017 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Computing differential atmospheric refraction"""

from __future__ import print_function

import astropy.units as u

import math

import numpy


@u.quantity_input(zenith_distance=u.micron)
def ref_index0(wl):
    wl0 = wl.to(u.micron).value
    wl1 = (1.0 / wl0)**2
    return 64.328 + 29498.1 / (146 - wl1) + 255.4 / (41 - wl1)


@u.quantity_input(zenith_distance=u.micron)
def ref_index1(wl, p=760, t=15):
    # 1 mmHg = 133.322387415
    n6_0 = ref_index0(wl)
    factor1 = p * (1 + (1.049 - 0.0157 * t) * 1e-6 * p)
    factor2 = 720.883 * (1 + 0.003661 * t)
    n6_1 = n6_0 * factor1 / factor2
    return n6_1


@u.quantity_input(zenith_distance=u.micron)
def ref_index2(wl, p=760, t=15, f=0.0):
    n6_1 = ref_index1(wl, p, t)
    wl0 = wl.to(u.micron).value
    wl1 = (1.0 / wl0) ** 2
    cont1 = 0.0624 - 0.000680 * wl1
    cont2 = 720.883 * (1 + 0.003661 * t)
    n6_2 = n6_1 - cont1 / cont2 * f
    return n6_2


@u.quantity_input(zenith_distance=u.deg)
def differential_p(
        zenith_distance,
        wl,
        wl_reference,
        temperature,
        pressure,
        relative_humidity,
):
    """Differential refraction as given by 1982PASP...94..715F """

    zd = zenith_distance.to(u.rad).value
    p = pressure.to(u.Pascal).value / 133.322387415
    t = temperature.to(u.deg_C, equivalencies=u.temperature()).value
    f = relative_humidity * p
    nl = 1 + 1e-6 * ref_index2(wl, p, t, f)
    n0 = 1 + 1e-6 * ref_index2(wl_reference, p, t, f)

    delt_r = (nl - n0) * math.tan(zd)

    return delt_r * u.rad


class DifferentialRefractionModel(object):
    def __init__(self, temperature, pressure, relative_humidity):
        self.temperature = temperature
        self.pressure = pressure
        self.relative_humidity = relative_humidity

    def refraction(self, z, wl, ref):
        return differential_p(z, wl, ref, self.temperature, self.pressure, self.relative_humidity)


if __name__ == '__main__':
    from astropy.units import cds
    cds.enable()

    wl = numpy.linspace(0.3, 1.0, num=15) * u.micron

    #wl = 0.6 * u.micron

    print(wl)
    zdis = 60*u.deg
    rel = 8.0 / 600.0
    ref = 0.5 * u.micron
    temp = 11.5 * u.deg_C
    press = 600 * cds.mmHg
    print(differential_p(zdis, wl, ref, temp, press, rel).to(u.arcsec))


# [ 0.3   0.35  0.4   0.45  0.5   0.55  0.6   0.65  0.7   0.75  0.8   0.85
#   0.9   0.95  1.  ] micron
# [ 139.21843488   75.42767344   39.3957048    16.18259071    0.
#   -11.86954952  -20.8895519   -27.92780533  -33.53521547  -38.07938831
#  -41.81507414  -44.92415491  -47.53967765  -49.76096379  -51.66345799] arcsec


#
# Copyright 2011-2019 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Dark current calibration Recipe for Megara"""

from numina.core import Result
from numina.array import combine

from megaradrp.core.recipe import MegaraBaseRecipe
from megaradrp.requirements import MasterBiasRequirement
from megaradrp.types import MasterDark
from megaradrp.processing.combine import basic_processing_with_combination


class DarkRecipe(MegaraBaseRecipe):

    """Process DARK images and provide MASTER_DARK."""

    master_bias = MasterBiasRequirement()

    master_dark = Result(MasterDark)

    def run(self, rinput):

        flow = self.init_filters(rinput, rinput.obresult.configuration)
        hdulist = basic_processing_with_combination(rinput, flow, method=combine.median)
        hdr = hdulist[0].header
        self.set_base_headers(hdr)

        result = self.create_result(master_dark=hdulist)
        return result

    def set_base_headers(self, hdr):
        """Set metadata in FITS headers."""
        hdr = super(DarkRecipe, self).set_base_headers(hdr)
        hdr['NUMTYPE'] = ('MasterDark', 'Product type')
        return hdr




import datetime

from numinadb.event import on_event

from .model import MegaraFrame


@on_event('on_ingest_raw_fits')
def function(session, some, meta):

    if meta['instrument'] != 'MEGARA':
        # only for MEGARA images
        return None

    res = session.query(MegaraFrame).filter_by(name=meta['path']).first()

    if res:
        # alreay inserted
        return res

    newframe = MegaraFrame()
    newframe.name = meta['path']
    newframe.uuid = meta['uuid']
    newframe.start_time = meta['observation_date']
    newframe.ob_id = meta['blckuuid']
    # No way of knowing when the readout ends...
    newframe.completion_time = newframe.start_time + datetime.timedelta(seconds=meta['darktime'])
    newframe.exposure_time = meta['exptime']
    newframe.object = meta['object']
    newframe.insmode = meta.get('insmode', 'unknown')
    newframe.vph = meta.get('vph', 'unknown')
    # ob.frames.append(newframe)
    # ob.object = meta['object']
    session.add(newframe)
    return newframe





def create_simple_hdul():
    """Create a simple image for testing"""
    import astropy.io.fits as fits

    prim = fits.PrimaryHDU()
    prim.header['instrume'] = 'MEGARA'
    prim.header['VPH'] = 'LR-B'
    prim.header['DATE-OBS'] = '2019-02-21T01:02:02.2'
    prim.header['insmode'] = 'LCB'
    prim.header['UUID'] = '410f6ea0-c3df-481c-9820-5cf9e0ed3d91'
    fibers = fits.ImageHDU(name='FIBERS')
    fibers.header['CONFID'] = 'a908e9d2-7599-41f3-9e9d-91042df01da5'
    simple_img = fits.HDUList([prim, fibers])
    return simple_img


#
# Copyright 2016-2019 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Combination routines"""



try:
    import contextlib2 as contextlib  # python 2.7 compatibility
except ImportError:
    import contextlib

from numina.array import combine
from numina.processing.combine import combine_imgs


def basic_processing_with_combination(
        rinput, reduction_flows,
        method=combine.mean, method_kwargs=None,
        errors=True, prolog=None):

    return basic_processing_with_combination_frames(
        rinput.obresult.frames, reduction_flows,
        method=method, method_kwargs=method_kwargs,
        errors=errors, prolog=prolog
    )


def basic_processing_with_combination_frames(
        frames, reduction_flows,
        method=combine.mean, method_kwargs=None,
        errors=True, prolog=None):
    """Perform basic reduction on set of DataFrames

    The reduction_flows are split in two parts.
    First part is performed in individual images (overscan and trimming)
    Then images are combined according to method and method_kwargs
    The resulting image is then processed with the
    second flow (bias, dark, gain and flat-fielding)
    """
    reduction_flow_ot, reduction_flow_1im = reduction_flows

    with contextlib.ExitStack() as stack:
        hduls = [stack.enter_context(dframe.open()) for dframe in frames]

        hdul_ot = [reduction_flow_ot(hdul) for hdul in hduls]

        hdu_combined = combine_imgs(hdul_ot, method=method, method_kwargs=method_kwargs,
                                    errors=errors, prolog=prolog)

        result = reduction_flow_1im(hdu_combined)

    return result


#
# Copyright 2011-2019 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

from __future__ import division, print_function

import uuid

import numpy
from scipy.ndimage.filters import median_filter
from astropy.io import fits
from numina.array import combine
from numina.core import Result, Parameter

from megaradrp.types import ProcessedFrame
from megaradrp.processing.combine import basic_processing_with_combination
from megaradrp.core.recipe import MegaraBaseRecipe
from megaradrp.types import MasterSlitFlat
import megaradrp.requirements as reqs
import megaradrp.core.correctors as cor


class SlitFlatRecipe(MegaraBaseRecipe):
    """Process SLIT_FLAT images and create MasterSlitFlat."""

    # Requirements
    master_bpm = reqs.MasterBPMRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()

    #window_length_x = Parameter(301, 'Savitzky-Golay length of the filter window OX')
    #window_length_y = Parameter(31, 'Savitzky-Golay length of the filter window OY')
    #polyorder = Parameter(3, 'Savitzky-Golay order of the polynomial used to fit the samples')
    #median_window_length = Parameter(31, 'Median window width')

    # Results
    reduced_image = Result(ProcessedFrame)
    master_slitflat = Result(MasterSlitFlat)

    def get_filters(self):
        return [cor.get_corrector_overscan, cor.get_corrector_trimming,
                cor.get_corrector_bpm, cor.get_corrector_bias,
                cor.get_corrector_dark, cor.get_corrector_gain]

    def run(self, rinput):
        from scipy.signal import savgol_filter

        self.logger.info('starting slit flat reduction')

        flow = self.init_filters(rinput, rinput.obresult.configuration)
        reduced = basic_processing_with_combination(rinput, flow, method=combine.median)
        hdr = reduced[0].header
        self.set_base_headers(hdr)

        self.save_intermediate_img(reduced, 'reduced_image.fits')

        # Using median filtering... In each channel
        l0 = reduced[0].data.shape[0]
        l1 = l0 // 2
        channel1 = (slice(None, l1), slice(None, None, None))
        channel2 = (slice(l1, None, None), slice(None, None, None))

        m_window1 = (11, 11)
        self.logger.debug('median filtering by channel %s', m_window1)
        median1 = numpy.zeros_like(reduced[0].data)
        median1[channel1] = median_filter(reduced[0].data[channel1], m_window1)
        median1[channel2] = median_filter(reduced[0].data[channel2], m_window1)

        self.save_intermediate_array(median1, 'median_image1.fits')

        qe1 = reduced[0].data / median1

        self.save_intermediate_array(qe1, 'qe_filter1.fits')

        m_window2 = (3, 21)
        self.logger.debug('median filtering %s', m_window2)

        median2 = median_filter(qe1, m_window2)
        self.save_intermediate_array(median2  , 'median_image2.fits')

        qe2 = qe1 / median2
        self.save_intermediate_array(qe2, 'qe_filter2.fits')

        self.logger.debug('filtering Inf/NaN in result')
        qe2[numpy.isinf(qe2)] = 1.0
        qe2[numpy.isnan(qe2)] = 1.0

        hdu = fits.PrimaryHDU(qe2.astype('float32'), header=reduced[0].header)
        hdu.header['UUID'] = str(uuid.uuid1())

        master_slitflat = fits.HDUList([hdu])
        self.set_base_headers(master_slitflat[0].header)

        self.logger.info('end slit flat recipe')
        return self.create_result(master_slitflat=master_slitflat,
                                  reduced_image=reduced)

    def set_base_headers(self, hdr):
        """Set metadata in FITS headers."""
        hdr = super(SlitFlatRecipe, self).set_base_headers(hdr)
        hdr['NUMTYPE'] = ('MasterSlitFlat', 'Product type')
        return hdr


#
# Copyright 2016-2019 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Data model for MEGARA"""

from __future__ import division

import re
import pkgutil
import enum

import astropy.io.fits as fits
import astropy.table
from six import StringIO
from numina.datamodel import DataModel, QueryAttribute, KeyDefinition
from numina.util.convert import convert_date

import megaradrp.instrument as megins


class MegaraDataModel(DataModel):
    """Data model of MEGARA images"""

    query_attrs = {
        'vph': QueryAttribute('vph', str),
        'insmode': QueryAttribute('insmode', str),
        'insconf': QueryAttribute('insconf', str),
        'speclamp': QueryAttribute('speclamp', str),
        'temp': QueryAttribute('temp', float),
        'confid': QueryAttribute('confid', str)
    }

    meta_info_headers = [
        'instrument',
        'object',
        'observation_date',
        'uuid',
        'type',
        'mode',
        'exptime',
        'darktime',
        'insconf',
        'blckuuid',
        'quality_control',
        'vph',
        'insmode'
    ]

    db_info_keys = [
        'instrument',
        'object',
        'observation_date',
        'uuid',
        'type',
        'mode',
        'exptime',
        'darktime',
        'insconf',
        'blckuuid',
        'quality_control',
        'vph',
        'insmode'
    ]

    db_info_keys_extra = [
        'vph',
        'insmode'
    ]

    meta_dinfo_headers = [
        'exptime',
        'observation_date',
        'vph',
        'vphpos',
        'insmode',
        'focus',
        'osfilter',
        'uuid',
        'temp',
        'block_uuid',
        'insconf_uuid',
        'speclamp',
        'imgid'
    ]

    PLATESCALE = megins.MEGARA_PLATESCALE

    def __init__(self):

        instrument_mappings = {
            'date_obs': ('DATE-OBS', 0, convert_date),
            'insconf': 'insconf',
            'insconf_uuid': 'insconf',
            'blckuuid': 'blckuuid',
            'block_uuid': 'blckuuid',
            'vph': ('VPH', 'undefined'),
            'vphpos': ('VPHWHPOS', 'undefined'),
            'focus': ('FOCUS', 'undefined'),
            'osfilter': ('OSFILTER', 'undefined'),
            'temp': ('SENTEMP4', 0.0),
            'speclamp': ('SPECLAMP', 'undefined'),
            'confid': KeyDefinition('CONFID', ext='FIBERS'),
        }
        super(MegaraDataModel, self).__init__(
            'MEGARA',
            instrument_mappings
        )

    def get_imgid(self, img):
        hdr = self.get_header(img)
        if 'UUID' in hdr:
            return 'uuid:{}'.format(hdr['UUID'])
        elif 'DATE-OBS' in hdr:
            return 'dateobs:{}'.format(hdr['DATE-OBS'])
        else:
            return super(MegaraDataModel, self).get_imgid(img)

    def get_fiberconf(self, img):
        """Obtain FiberConf from image"""
        main_insmode = img[0].header.get('INSMODE', '')
        if 'FIBERS' in img:
            # We have a 'fibers' extension
            # Information os there
            hdr_fiber = img['FIBERS'].header
            return read_fibers_extension(hdr_fiber, insmode=main_insmode)
        else:
            return self.get_fiberconf_default(main_insmode)

    def get_fiberconf_default(self, insmode):
        """Obtain default FiberConf object"""
        if insmode == 'LCB':
            slit_file = 'lcb_default_header.txt'
        elif insmode == 'MOS':
            slit_file = 'mos_default_header.txt'
        else:
            # Read fiber info from headers
            raise ValueError('Invalid INSMODE {}'.format(insmode))

        data = pkgutil.get_data('megaradrp.instrument.configs', slit_file)
        default_hdr = StringIO(data.decode('utf8'))
        hdr_fiber = fits.header.Header.fromfile(default_hdr)
        return read_fibers_extension(hdr_fiber)

    def gather_info_oresult(self, val):
        return [self.gather_info_dframe(f) for f in val.images]

    def fiber_scale_unit(self, img, unit=False):
        funit = img['FIBERS'].header.get("FUNIT", "arcsec")

        if funit == "arcsec":
            scale = 1
        else:
            scale = self.PLATESCALE
        if unit:
            return (scale, funit)
        else:
            return scale


class FibersConf(object):
    """Global configuration of the fibers"""
    def __init__(self):
        self.name = ""
        self.conf_id = 1
        self.nbundles = 0
        self.nfibers = 0
        self.bundles = {}
        self.fibers = {}
        self.funit = "mm"

    def sky_fibers(self, valid_only=False, ignored_bundles=None):
        result = []
        if ignored_bundles is None:
            ignored_bundles = []

        for bundle in self.bundles.values():
            if bundle.id in ignored_bundles:
                continue
            if bundle.target_type is TargetType.SKY:
                if valid_only:
                    for fib in bundle.fibers.values():
                        if fib.valid:
                            result.append(fib.fibid)
                else:
                    result.extend(bundle.fibers.keys())
        return result

    def conected_fibers(self, valid_only=False):

        if self.name == 'MOS':
            raise ValueError('not working for MOS')

        result = []
        for bundle in self.bundles.values():
            if bundle.target_type is not TargetType.SKY:
                if valid_only:
                    for fib in bundle.fibers.values():
                        if fib.valid:
                            result.append(fib)
                else:
                    result.extend(bundle.fibers.values())
        return result

    def inactive_fibers(self):
        result = []
        for fiber in self.fibers.values():
            if fiber.inactive:
                result.append(fiber.fibid)
        return result

    def active_fibers(self):
        result = []
        for fiber in self.fibers.values():
            if not fiber.inactive:
                result.append(fiber.fibid)
        return result

    def valid_fibers(self):
        result = []
        for fiber in self.fibers.values():
            if fiber.valid:
                result.append(fiber.fibid)
        return result

    def invalid_fibers(self):
        result = []
        for fiber in self.fibers.values():
            if not fiber.valid:
                result.append(fiber.fibid)
        return result

    def spectral_coverage(self):
        lowc = []
        upperc = []
        for fibid, r in self.fibers.items():
            if r.w1:
                lowc.append(r.w1)
            if r.w2:
                upperc.append(r.w2)

        mn = max(lowc)
        nn = min(lowc)

        mx = min(upperc)
        nx = max(upperc)
        return (mn, mx), (nn, nx)

    def bundles_to_table(self):
        """Convert bundles to a Table"""
        attrnames = ['id', 'x', 'y', 'pa', 'enabled',
                  'target_type', 'target_priority', 'target_name']
        cnames = ['bundle_id', 'x', 'y', 'pa', 'enabled',
                  'target_type', 'target_priority', 'target_name']
        obj_data = {}
        for a, c in zip(attrnames, cnames):
            obj_data[c] = [getattr(ob, a) for ob in self.bundles.values()]
        result = astropy.table.Table(obj_data, names=cnames)
        result['x'].unit = self.funit
        result['y'].unit = self.funit
        result['pa'].unit = 'deg'
        return result

    def fibers_to_table(self):
        """Convert fibers to a Table"""
        attrnames = ['fibid', 'name', 'x', 'y', 'inactive', 'valid',
                     'bundle_id']
        cnames = ['fibid', 'name', 'x', 'y', 'inactive', 'valid',
                  'bundle_id']
        obj_data = {}

        for a, c in zip(attrnames, cnames):
            obj_data[c] = [getattr(ob, a) for ob in self.fibers.values()]
        result = astropy.table.Table(obj_data, names=cnames)
        result['x'].unit = self.funit
        result['y'].unit = self.funit
        return result



class TargetType(enum.Enum):
    """Possible targest in a fiber bundle"""
    SOURCE = 1
    UNKNOWN = 2
    UNASSIGNED = 3
    SKY = 4
    REFERENCE = 5
    # aliases for the other fields
    STAR = 5
    BLANK = 4


class BundleConf(object):
    """Description of a bundle"""
    def __init__(self):
        self.id = 0
        self.target_type = TargetType.UNASSIGNED
        self.target_priority = 0
        self.target_name = 'unknown'
        self.x_fix = 0
        self.y_fix = 0
        self.pa_fix = 0
        self.x = 0
        self.y = 0
        self.pa = 0
        self.enabled = True


class FiberConf(object):
    """Description of the fiber"""
    def __init__(self):
        self.fibid = 0
        self.name = 'unknown'
        self.bundle_id = None
        self.inactive = False
        self.valid = True
        self.x = 0.0
        self.y = 0.0


def read_fibers_extension(hdr, insmode='LCB'):
    """Read the FIBERS extension

    Parameters
    ==========
    hdr:
       FITS header
    insmode: str
        default INSMODE

    Returns
    =======
    FibersConf


    """
    conf = FibersConf()
    defaults = {}
    defaults['LCB'] = (9, 623)
    defaults['MOS'] = (92, 644)

    if insmode not in ['LCB', 'MOS']:
        raise ValueError('insmode %s not in [LCB, MOS]' % insmode)

    conf.name = hdr.get('INSMODE', insmode)
    conf.conf_id = hdr.get('CONFID', 1)
    conf.nbundles = hdr.get('NBUNDLES', defaults[insmode][0])
    conf.nfibers = hdr.get('NFIBERS', defaults[insmode][1])
    conf.funit = funit = hdr.get("FUNIT", "arcsec")
    # Read bundles

    bun_ids = []
    fib_ids = []
    bundles = conf.bundles
    fibers = conf.fibers

    # loop over everything, count BUN%03d_P and FIB%03d_B
    pattern1 = re.compile(r"BUN(\d+)_P")
    pattern2 = re.compile(r"FIB(\d+)_B")
    for key in hdr:
        bun_re = pattern1.match(key)
        fib_re = pattern2.match(key)
        if bun_re:
            bun_idx = int(bun_re.group(1))
            bun_ids.append(bun_idx)
        elif fib_re:
            fib_idx = int(fib_re.group(1))
            fib_ids.append(fib_idx)

    for i in bun_ids:
        bb = BundleConf()
        bb.id = i
        bb.target_priority = hdr["BUN%03d_P" % i]
        bb.target_name = hdr["BUN%03d_I" % i]
        bb.target_type = TargetType[hdr["BUN%03d_T" % i]]
        bb.enabled = hdr.get("BUN%03d_E" % i, True)
        bb.x = hdr.get("BUN%03d_X" % i, 0.0)
        bb.y = hdr.get("BUN%03d_Y" % i, 0.0)
        bb.pa = hdr.get("BUN%03d_O" % i, 0.0)
        bb.fibers = {}
        bundles[i] = bb

    for fibid in fib_ids:
        ff = FiberConf()
        ff.fibid = fibid

        # Coordinates
        ff.d = hdr["FIB%03d_D" % fibid]
        ff.r = hdr["FIB%03d_R" % fibid]
        ff.o = 0 #hdr["FIB%03d_O" % fibid]
        # Active
        ff.inactive = not hdr["FIB%03d_A" % fibid]

        # Coordinates XY
        ff.x = hdr["FIB%03d_X" % fibid]
        ff.y = hdr["FIB%03d_Y" % fibid]

        ff.bundle_id = hdr["FIB%03d_B" % fibid]
        ff.name = hdr.get("FIB%03d_N" % fibid, 'unknown')

        ff.w1 = hdr.get("FIB%03dW1" % fibid, None)
        ff.w2 = hdr.get("FIB%03dW2" % fibid, None)

        # Validity
        if ff.inactive:
            ff.valid = False
        else:
            ff.valid = hdr.get("FIB%03d_V" % fibid, True)

        bundles[ff.bundle_id].fibers[ff.fibid] = ff
        fibers[ff.fibid] = ff

    return conf
